/* ***********************************************
 
	CSRMV-N for CUDA 

   *********************************************** */   

#include "mublas_csrmv_n.def"

/* ***********************************************
    Macro
   *********************************************** */   
#if USE_NC_csrValA_CSRMV_N
#define RESTRICT_csrValA __restrict__
#define LOAD1_csrValA LOAD1_NC
#else
#define RESTRICT_csrValA
#define LOAD1_csrValA LOAD1___
#endif

#if USE_NC_csrRowPtrA_CSRMV_N
#define RESTRICT_csrRowPtrA __restrict__
#define LOAD1_csrRowPtrA LOAD1_NC
#else
#define RESTRICT_csrRowPtrA
#define LOAD1_csrRowPtrA LOAD1___
#endif

#if USE_NC_csrColIndA_CSRMV_N
#define RESTRICT_csrColIndA __restrict__
#define LOAD1_csrColIndA LOAD1_NC
#else
#define RESTRICT_csrColIndA
#define LOAD1_csrColIndA LOAD1___
#endif

#if USE_NC_X_CSRMV_N
#define RESTRICT_X __restrict__
#define LOAD1_X LOAD1_NC
#else
#define RESTRICT_X
#define LOAD1_X LOAD1___
#endif

#define LOAD1_Y LOAD1___
#define STORE1_Y STORE1___

/* ***********************************************
	Functions for determining Tx and Ty
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_CSRMV_N (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	mh->Bx = mh->m / (mh->Tx/mh->N_T) + ((mh->m % (mh->Tx/mh->N_T)) != 0);
	mh->By = 1;
}

__host__ __inline__ void getShmPerBlk_CSRMV_N (void* mh_) {
	mublasHandle_t* mh = (mublasHandle_t*) mh_;
	if (mh->cc >= 300)
		mh->ShmPerBlk = 0.;
	else
		mh->ShmPerBlk = ceilingEx ((float)(WrpSize * mh->N_one * mh->N_T), mh->ShmUnit);
}

__host__ __inline__ size_t getTxSuf_CSRMV_N (mublasHandle_t* mh) {
	return (size_t)ceilingEx ((float)mh->n / (float)mh->devprop.multiProcessorCount, mh->TxStep) * WrpSize;
}

/* ***********************************************
	kernel=s1, trans=N
   *********************************************** */   
__global__ void MUBLAS_CSRMV_N_DEVICE_N32 (
	const SIZE_T m,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_csrValA csrValA,
	const int32_t * RESTRICT_csrRowPtrA csrRowPtrA,
	const int32_t * RESTRICT_csrColIndA csrColIndA,
	const TYPE0 * RESTRICT_X x,
	const TYPE0 beta,
	TYPE0 *y
) {
	const SIZE_T tx = threadIdx.x;
	const SIZE_T tid = blockDim.x * blockIdx.x + tx;
	const SIZE_T rowid = tid / 32;
	const SIZE_T lane  = tid % 32; 
	SIZE_T i;
	TYPE1 Tr1, Xr1, Ar1, Yr1;
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
#else
	extern __shared__ TYPE1 Ts1[];
#endif

	if (rowid < m){
		Tr1 = ZERO1;
		for (i = csrRowPtrA[rowid] + lane; i < csrRowPtrA[rowid+1]; i += 32) {
			LOAD1_csrValA (Ar1, &csrValA[i]);
			LOAD1_X (Xr1, &x[csrColIndA[i]]);
			Tr1 = MAD1 (Ar1, Xr1, Tr1);
		}

		// reduction
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
		#pragma unroll
		for (i = 16; i > 0; i >>= 1) 
			shuffleAdd (Tr1, i);
#else
		Ts1[tx] = Tr1;
		if (lane < 16) Ts1[tx] = ADD1 (Ts1[tx+16], Ts1[tx]);
		if (lane <  8) Ts1[tx] = ADD1 (Ts1[tx+ 8], Ts1[tx]);
		if (lane <  4) Ts1[tx] = ADD1 (Ts1[tx+ 4], Ts1[tx]);
		if (lane <  2) Ts1[tx] = ADD1 (Ts1[tx+ 2], Ts1[tx]);
		if (lane <  1) Tr1 = ADD1 (Ts1[tx+ 1], Ts1[tx]);
#endif

		if (lane == 0) {
			LOAD1_Y (Yr1, &y[rowid]);
			Yr1 = MAD1 (LOAD0(alpha), Tr1, MUL1 (LOAD0(beta), Yr1));
			STORE1_Y (&y[rowid], Yr1);
		}
	}
}

__global__ void MUBLAS_CSRMV_N_DEVICE_N16 (
	const SIZE_T m,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_csrValA csrValA,
	const int32_t * RESTRICT_csrRowPtrA csrRowPtrA,
	const int32_t * RESTRICT_csrColIndA csrColIndA,
	const TYPE0 * RESTRICT_X x,
	const TYPE0 beta,
	TYPE0 *y
) {
	const SIZE_T tx = threadIdx.x;
	const SIZE_T tid = blockDim.x * blockIdx.x + tx;
	const SIZE_T rowid = tid / 16;
	const SIZE_T lane  = tid % 16; 
	SIZE_T i;
	TYPE1 Tr1, Xr1, Ar1, Yr1;
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
#else
	extern __shared__ TYPE1 Ts1[];
#endif

	if (rowid < m){
		Tr1 = ZERO1;
		for (i = csrRowPtrA[rowid] + lane; i < csrRowPtrA[rowid+1]; i += 16) {
			LOAD1_csrValA (Ar1, &csrValA[i]);
			LOAD1_X (Xr1, &x[csrColIndA[i]]);
			Tr1 = MAD1 (Ar1, Xr1, Tr1);
		}

		// reduction
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
		#pragma unroll
		for (i = 8; i > 0; i >>= 1) 
			shuffleAdd (Tr1, i);
#else
		Ts1[tx] = Tr1;
		if (lane <  8) Ts1[tx] = ADD1 (Ts1[tx+ 8], Ts1[tx]);
		if (lane <  4) Ts1[tx] = ADD1 (Ts1[tx+ 4], Ts1[tx]);
		if (lane <  2) Ts1[tx] = ADD1 (Ts1[tx+ 2], Ts1[tx]);
		if (lane <  1) Tr1 = ADD1 (Ts1[tx+ 1], Ts1[tx]);
#endif

		if (lane == 0) {
			LOAD1_Y (Yr1, &y[rowid]);
			Yr1 = MAD1 (LOAD0(alpha), Tr1, MUL1 (LOAD0(beta), Yr1));
			STORE1_Y (&y[rowid], Yr1);
		}
	}
}

__global__ void MUBLAS_CSRMV_N_DEVICE_N8 (
	const SIZE_T m,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_csrValA csrValA,
	const int32_t * RESTRICT_csrRowPtrA csrRowPtrA,
	const int32_t * RESTRICT_csrColIndA csrColIndA,
	const TYPE0 * RESTRICT_X x,
	const TYPE0 beta,
	TYPE0 *y
) {
	const SIZE_T tx = threadIdx.x;
	const SIZE_T tid = blockDim.x * blockIdx.x + tx;
	const SIZE_T rowid = tid / 8;
	const SIZE_T lane  = tid % 8; 
	SIZE_T i;
	TYPE1 Tr1, Xr1, Ar1, Yr1;
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
#else
	extern __shared__ TYPE1 Ts1[];
#endif

	if (rowid < m){
		Tr1 = ZERO1;
		for (i = csrRowPtrA[rowid] + lane; i < csrRowPtrA[rowid+1]; i += 8) {
			LOAD1_csrValA (Ar1, &csrValA[i]);
			LOAD1_X (Xr1, &x[csrColIndA[i]]);
			Tr1 = MAD1 (Ar1, Xr1, Tr1);
		}

		// reduction
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
		#pragma unroll
		for (i = 4; i > 0; i >>= 1) 
			shuffleAdd (Tr1, i);
#else
		Ts1[tx] = Tr1;
		if (lane <  4) Ts1[tx] = ADD1 (Ts1[tx+ 4], Ts1[tx]);
		if (lane <  2) Ts1[tx] = ADD1 (Ts1[tx+ 2], Ts1[tx]);
		if (lane <  1) Tr1 = ADD1 (Ts1[tx+ 1], Ts1[tx]);
#endif

		if (lane == 0) {
			LOAD1_Y (Yr1, &y[rowid]);
			Yr1 = MAD1 (LOAD0(alpha), Tr1, MUL1 (LOAD0(beta), Yr1));
			STORE1_Y (&y[rowid], Yr1);
		}
	}
}

__global__ void MUBLAS_CSRMV_N_DEVICE_N4 (
	const SIZE_T m,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_csrValA csrValA,
	const int32_t * RESTRICT_csrRowPtrA csrRowPtrA,
	const int32_t * RESTRICT_csrColIndA csrColIndA,
	const TYPE0 * RESTRICT_X x,
	const TYPE0 beta,
	TYPE0 *y
) {
	const SIZE_T tx = threadIdx.x;
	const SIZE_T tid = blockDim.x * blockIdx.x + tx;
	const SIZE_T rowid = tid / 4;
	const SIZE_T lane  = tid % 4; 
	SIZE_T i;
	TYPE1 Tr1, Xr1, Ar1, Yr1;
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
#else
	extern __shared__ TYPE1 Ts1[];
#endif

	if (rowid < m){
		Tr1 = ZERO1;
		for (i = csrRowPtrA[rowid] + lane; i < csrRowPtrA[rowid+1]; i += 4) {
			LOAD1_csrValA (Ar1, &csrValA[i]);
			LOAD1_X (Xr1, &x[csrColIndA[i]]);
			Tr1 = MAD1 (Ar1, Xr1, Tr1);
		}

		// reduction
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
		#pragma unroll
		for (i = 2; i > 0; i >>= 1) 
			shuffleAdd (Tr1, i);
#else
		Ts1[tx] = Tr1;
		if (lane <  2) Ts1[tx] = ADD1 (Ts1[tx+ 2], Ts1[tx]);
		if (lane <  1) Tr1 = ADD1 (Ts1[tx+ 1], Ts1[tx]);
#endif

		if (lane == 0) {
			LOAD1_Y (Yr1, &y[rowid]);
			Yr1 = MAD1 (LOAD0(alpha), Tr1, MUL1 (LOAD0(beta), Yr1));
			STORE1_Y (&y[rowid], Yr1);
		}
	}
}

__global__ void MUBLAS_CSRMV_N_DEVICE_N2 (
	const SIZE_T m,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_csrValA csrValA,
	const int32_t * RESTRICT_csrRowPtrA csrRowPtrA,
	const int32_t * RESTRICT_csrColIndA csrColIndA,
	const TYPE0 * RESTRICT_X x,
	const TYPE0 beta,
	TYPE0 *y
) {
	const SIZE_T tx = threadIdx.x;
	const SIZE_T tid = blockDim.x * blockIdx.x + tx;
	const SIZE_T rowid = tid / 2;
	const SIZE_T lane  = tid % 2; 
	SIZE_T i;
	TYPE1 Tr1, Xr1, Ar1, Yr1;
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
#else
	extern __shared__ TYPE1 Ts1[];
#endif

	if (rowid < m){
		Tr1 = ZERO1;
		for (i = csrRowPtrA[rowid] + lane; i < csrRowPtrA[rowid+1]; i += 2) {
			LOAD1_csrValA (Ar1, &csrValA[i]);
			LOAD1_X (Xr1, &x[csrColIndA[i]]);
			Tr1 = MAD1 (Ar1, Xr1, Tr1);
		}

		// reduction
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
		#pragma unroll
		for (i = 1; i > 0; i >>= 1) 
			shuffleAdd (Tr1, i);
#else
		Ts1[tx] = Tr1;
		if (lane <  1) Tr1 = ADD1 (Ts1[tx+ 1], Ts1[tx]);
#endif

		if (lane == 0) {
			LOAD1_Y (Yr1, &y[rowid]);
			Yr1 = MAD1 (LOAD0(alpha), Tr1, MUL1 (LOAD0(beta), Yr1));
			STORE1_Y (&y[rowid], Yr1);
		}
	}
}

__global__ void MUBLAS_CSRMV_N_DEVICE_N1 (
	const SIZE_T m,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_csrValA csrValA,
	const int32_t * RESTRICT_csrRowPtrA csrRowPtrA,
	const int32_t * RESTRICT_csrColIndA csrColIndA,
	const TYPE0 * RESTRICT_X x,
	const TYPE0 beta,
	TYPE0 *y
) {
	const SIZE_T tx = threadIdx.x;
	const SIZE_T tid = blockDim.x * blockIdx.x + tx;
	const SIZE_T rowid = tid;
	SIZE_T i;
	TYPE1 Tr1, Xr1, Ar1, Yr1;

	if (rowid < m){
		Tr1 = ZERO1;
		for (i = csrRowPtrA[rowid]; i < csrRowPtrA[rowid+1]; i += 1) {
			LOAD1_csrValA (Ar1, &csrValA[i]);
			LOAD1_X (Xr1, &x[csrColIndA[i]]);
			Tr1 = MAD1 (Ar1, Xr1, Tr1);
		}
		LOAD1_Y (Yr1, &y[rowid]);
		Yr1 = MAD1 (LOAD0(alpha), Tr1, MUL1 (LOAD0(beta), Yr1));
		STORE1_Y (&y[rowid], Yr1);
	}
}

#undef RESTRICT_csrValA
#undef RESTRICT_csrRowPtrA
#undef RESTRICT_csrColIndA
#undef RESTRICT_X
#undef LOAD1_csrValA
#undef LOAD1_csrRowPtrA
#undef LOAD1_csrColIndA
#undef LOAD1_X
#undef LOAD1_Y
#undef STORE1_Y

