/* ***********************************************
 
	GEMV for CUDA 

   *********************************************** */   

#include "mublas_devel.h"
#include "mublas_gemv_n_kernel.cu"
#include "mublas_gemv_t_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   

__host__ int32_t
MUBLAS_GEMV (
	mublasHandle_t* mh,	
	const char trans,
	const size_t m,
	const size_t n,
	const TYPE0 *alpha,
	const TYPE0 *a,
	const size_t lda,
	const TYPE0 *x,
	const size_t incx,
	const TYPE0 *beta,
	TYPE0 *y,
	const size_t incy
) {
	uint32_t kernel_number = 999;
	uint32_t nonvectorkernel = 0;
	void (*kernel) (const SIZE_T, const SIZE_T, const TYPE0, const TYPE0*, const SIZE_T, const TYPE0*, const TYPE0, TYPE0 *);
	mublasRecipeDefault (mh);

	// ---------------------------------------------
	// 1. Shortcut
	if (m == 0 || n == 0 || (*alpha == ZERO0 && *beta == ONE0 )) 
		return 0;
	if (*alpha == ZERO0) {
		MUBLAS_SCAL (mh, m, beta, y, incy);
		return 0;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration 
	mh->m = m;
	mh->n = n;
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	if ((((long)a%(128/sizeof(TYPE0)))%16 != 0) || 
		(((long)x%(128/sizeof(TYPE0)))%16 != 0) ||
		(((long)y%(128/sizeof(TYPE0)))%16 != 0) ||
		(lda%mh->N_T != 0)) nonvectorkernel = 1;
	if (incx == 1 && incy == 1) {
		if (trans == 'N' || trans == 'n') {
			mh->BlkDim = 2;
			mh->func_getShmPerBlk = &getShmPerBlk_GEMV_N;
			mh->func_getBlkPerGrd = &getBlkPerGrd_GEMV_N;
			mh->func_cache = cudaFuncCachePreferShared;
			#include "mublas_gemv_n.incl"
			kernel_number = 100 + mh->N_T * 10 + nonvectorkernel;
		} else if (trans == 'T' || trans == 't') {
			mh->BlkDim = 1;
			mh->func_getShmPerBlk = &getShmPerBlk_GEMV_T;
			mh->func_getBlkPerGrd = &getBlkPerGrd_GEMV_T;
			mh->func_cache = cudaFuncCachePreferShared;
			#include "mublas_gemv_t.incl"
			kernel_number = 200 + mh->N_T * 10 + nonvectorkernel;
		}
	}
	switch (kernel_number) {
		case 140: kernel = &MUBLAS_GEMV_N_DEVICE_v4; break;
		case 141: kernel = &MUBLAS_GEMV_N_DEVICE_s4; break;
		case 120: kernel = &MUBLAS_GEMV_N_DEVICE_v2; break;
		case 121: kernel = &MUBLAS_GEMV_N_DEVICE_s2; break;
		case 110: kernel = &MUBLAS_GEMV_N_DEVICE_s1; break;
		case 240: kernel = &MUBLAS_GEMV_T_DEVICE_v4; break;
		case 220: kernel = &MUBLAS_GEMV_T_DEVICE_v2; break;
		case 210: kernel = &MUBLAS_GEMV_T_DEVICE_s1; break;
	}
	// ---------------------------------------------
	
	// ---------------------------------------------
	// 3. Unsupported operations (CUBLAS)
	if (kernel_number == 999) {
		#if defined (CUBLAS_GEMV)
		cublasHandle_t ch;
		cublasCreate (&ch);
		cublasSetStream (ch, mh->streamId);
		cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
		cublasOperation_t tranAcublas = CUBLAS_OP_N;
		if (trans == 'T' || trans == 't') tranAcublas = CUBLAS_OP_T;
		if (trans == 'C' || trans == 'c') tranAcublas = CUBLAS_OP_C;
		CUBLAS_GEMV (ch, tranAcublas, m, n, alpha, a, lda, x, incx, beta, y, incy);
		cublasDestroy (ch);
		#else
		printf ("mublas error: not implemented");
		#endif
		mh->kernelnumber = kernel_number;
		return 0;
	}
	// ---------------------------------------------
	
	// ---------------------------------------------
	// 4. Thread-Block Size Adjustment and Kernel Launch
	cudaFuncGetAttributes (&mh->funcatb, kernel);
	cudaFuncSetCacheConfig (kernel, mh->func_cache);
	mublas_optimizer (mh);
	dim3 threads = dim3 (mh->Tx, mh->Ty);
	dim3 grid = dim3 (mh->Bx, mh->By);
	if (trans == 'N' || trans == 'n') {
		if (mh->autotuning!=1) kernel <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
	} else {
		if (mh->autotuning!=1) kernel <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
	}
	// ---------------------------------------------

	mh->kernelnumber = kernel_number;
	return 0;
}

