// --------------------------------------------
// Power performance evaluation using NVML
// Oct. 4, 2016, Daichi Mukunoki
// --------------------------------------------

#include "testing_common.h"
#define PREC_NO
#include <sys/wait.h>
#include "testing_common.cpp"
#include <gsl/gsl_integration.h>
#include <gsl/gsl_spline.h>

// buffer size
#define WBUFFER 10000

typedef struct {
	gsl_interp_accel *acc;
	gsl_spline *spline;
}param_t;

// GSL function
double func (const double x, void *param_v) {
	param_t *param = (param_t*)param_v;
	gsl_interp_accel *acc = param->acc;
	gsl_spline *spline = param->spline;
	return gsl_spline_eval (spline, x, acc);
}

int32_t
main (int32_t argc, char **argv)
{
	pid_t pid;
	FILE *fp;
	char buf[256], command[512];
	int32_t  status, count, exec_iter;
	uint32_t *power, i, igpu, numgpus;
	size_t mm, nn, kk;
	double *powers, times[WBUFFER];
	double *gsl_rets, *gsl_errs;
	double x0, x1, exec_time, exec_flops, exec_gbs;

// ----- testing setup ------------------------
	testingHandle_t th;
	testingCreate (argc, argv, &th);
	i = 2;
	strcat (command, th.exec);
	strcat (command, " --nodisp=1 --mode=e");
	while (argv[i]) {
		strcat (command, " ");
		strcat (command, argv[i++]);
	}
	strcat (command, " > test.dat_");
	numgpus = th.num_installed_gpus;
// ---------------------------------------------

// malloc
	power = (uint32_t *) malloc (sizeof(uint32_t)*numgpus);
	powers = (double *) malloc (sizeof(double)*(numgpus+1)*WBUFFER);
	gsl_rets = (double *) malloc (sizeof(double)*(numgpus+1));
	gsl_errs = (double *) malloc (sizeof(double)*(numgpus+1));

// ----- NVML setup & evaluation ---------------
	nvmlInit();
	nvmlDevice_t* nvmldevs;
	nvmldevs = (nvmlDevice_t *)malloc(sizeof(nvmlDevice_t)*numgpus);
	for (igpu = 0; igpu < numgpus; igpu++) 
		nvmlDeviceGetHandleByIndex(igpu, &nvmldevs[igpu]);

	for (igpu = 0; igpu < numgpus+1; igpu++)
		for (count = 0; count < WBUFFER; count++)
			powers[igpu*WBUFFER+count] = 0.;

	count = 0;
	if ((pid = fork ()) < 0)
		perror ("fork error");
    else if (pid == 0) {
		system (command);
		exit (1);
	} else do {
		if ((pid = waitpid (pid, &status, WNOHANG)) == -1)
			perror ("waitpid error");
		else if (pid == 0) {
// ----- evaluation ----------------------------
			times[count] = gettime2();
			for (igpu = 0; igpu < numgpus; igpu++)
				nvmlDeviceGetPowerUsage (nvmldevs[igpu], &power[igpu]);
			printf ("%d\t%f\t%f\t", count, times[count], gettime2());
			for (igpu = 0; igpu < numgpus; igpu++) {
				powers[igpu*WBUFFER+count] = (double)power[igpu]/1000;
				powers[numgpus*WBUFFER+count] += powers[igpu*WBUFFER+count];
				printf ("%10.4f\t", powers[igpu*WBUFFER+count]);
			}
			printf ("%10.4f\n", powers[numgpus*WBUFFER+count]);
			count++;
			usleep (15000); //15msec
// ---------------------------------------------
		} else {
			// exit
		}
	}
    while (pid == 0);
	nvmlShutdown();

	if ((fp = fopen ("test.dat_", "r")) == NULL) {
		printf ("mublas testing error: file open error.\n");
		exit (1);
	}	
	while (fgets (buf, 256, fp) != NULL) 
		sscanf (buf, "%zd\t%zd\t%zd\t%lf\t%lf\t%le\t%le\t%le\t%d", &mm, &nn, &kk, &x0, &x1, &exec_time, &exec_flops, &exec_gbs, &exec_iter);
	fclose (fp);
	exec_time = x1 - x0; // exec_time from buf is the time for one execution
// ---------------------------------------------

// ----- GSL integration -----------------------
	gsl_interp_accel *acc = gsl_interp_accel_alloc();
	gsl_spline *spline = gsl_spline_alloc(gsl_interp_cspline, count);
	gsl_integration_workspace * work = gsl_integration_workspace_alloc (WBUFFER);
	gsl_function Func;
	for (igpu = 0; igpu < numgpus+1; igpu++) {
		gsl_spline_init (spline, times, &powers[igpu*WBUFFER], count);
		param_t param = {acc, spline};
		Func.function = &func;
		Func.params = &param;
		gsl_integration_qag (&Func, x0, x1, 0, 1.0e-4, WBUFFER, GSL_INTEG_GAUSS41, work, &gsl_rets[igpu], &gsl_errs[igpu]);
	}
	gsl_integration_workspace_free(work);
	gsl_spline_free(spline);
	gsl_interp_accel_free(acc);
// ---------------------------------------------

// results
	printf ("%% M\tN\tK\tTime[sec]\t#iter\t[GFlops]\t");
	for (igpu = 0; igpu < numgpus; igpu++) printf ("GPU-%d:[W]\t",igpu);
	printf ("GPU-total:[W]\t");
	printf ("GPU-%d:[GFlops/W]\t",th.devno);
	printf ("GPU-total:[GFlops/W]\n");
	printf ("%zd\t%zd\t%zd\t%1.3e\t%d\t%1.3e\t", mm, nn, kk, exec_time, exec_iter, exec_flops);
	for (igpu = 0; igpu < numgpus; igpu++) printf ("%1.3e\t", gsl_rets[igpu]/exec_time);
	printf ("%1.3e\t", gsl_rets[numgpus]/exec_time);
	printf ("%1.3e\t", exec_flops/(gsl_rets[th.devno]/exec_time));
	printf ("%1.3e\n", exec_flops/(gsl_rets[numgpus]/exec_time));

// free
	free (gsl_rets);
	free (gsl_errs);
	free (power);
	free (powers);
}

